@@ -32,6 +32,10 @@ perl_huffman_encode (SV * in)
     int rc;
     SV * RETVAL;
     cin = (unsigned char *) SvPV (in, cinlen);
+    if (cinlen == 0) {
+	warn ("Empty input");
+	return & PL_sv_undef;
+    }
     rc = huffman_encode_memory (cin, (unsigned int) cinlen, & out, & outlen);
     if (rc != 0) {
 	croak ("Error encoding scalar");
@@ -57,6 +61,10 @@ perl_huffman_decode (SV * in)
     SV * RETVAL;
 
     cin = (unsigned char *) SvPV (in, cinlen);
+    if (cinlen == 0) {
+	warn ("Empty input");
+	return & PL_sv_undef;
+    }
     rc = huffman_decode_memory (cin, (unsigned int) cinlen, & out, & outlen);
     if (rc != 0) {
 	croak ("Error decoding scalar");
