@@ -1,18 +1,16 @@
-/*
- *  huffman - Encode/Decode files using Huffman encoding.
- *  http://huffman.sourceforge.net
- *  Copyright (C) 2003  Douglas Ryan Richardson
- */
-
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <stdint.h>
 #include "huffman.h"
 
+/* The following #include gets ntohl and htonl on Windows and
+   Unix. ntohl and htonl are functions for converting uint32_t from
+   network (n) to host (h) order. These are used for storing binary
+   data on the disc in a system-independent way. */
+
 #ifdef WIN32
 #include <winsock2.h>
-#include <malloc.h>
-#define alloca _alloca
 #else
 #include <netinet/in.h>
 #endif
@@ -27,17 +25,30 @@
 #define NOIGNORE
 #endif /* def __GNUC__ */
 
+/* Return statuses of functions in this file. */
+
 typedef enum 
 {
     huffman_status_ok,
     huffman_status_memory_failure,
     huffman_status_null_pointer,
     huffman_status_bounds,
+    huffman_status_read_error,
 }
 huffman_status_t;
 
 #endif /* def HEADER */
 
+#define CALL(x) {							\
+	huffman_status_t rc;						\
+	rc = x;								\
+	if (rc != huffman_status_ok) {					\
+	    fprintf (stderr, "%s:%d: %s failed with status %d.\n",	\
+		     __FILE__, __LINE__, #x, rc);			\
+	    return rc;							\
+	}								\
+    }
+
 /* Count mallocs and frees. */
 
 //#define MEMCOUNT
@@ -78,10 +89,6 @@ static int n_mallocs;
 
 typedef struct huffman_node_tag
 {
-    /* One for a leaf node. */
-    unsigned char isLeaf;
-    /* Zero for a leaf node. */
-    unsigned long count;
     struct huffman_node_tag *parent;
 
     union
@@ -90,8 +97,12 @@ typedef struct huffman_node_tag
 	{
 	    struct huffman_node_tag *zero, *one;
 	};
-	unsigned char symbol;
+	unsigned int symbol;
     };
+    /* Zero for a leaf node. */
+    double count;
+    /* One for a leaf node. */
+    unsigned int is_leaf : 1;
 }
 huffman_node;
 
@@ -128,46 +139,48 @@ get_bit (unsigned char* bits, unsigned long i)
     return (bits[i / 8] >> i % 8) & 1;
 }
 
-/* Reverse "bits" in place. */
+/* Reverse "numbits" bits of "bits" in place. */
 
-static void
+static huffman_status_t NOIGNORE
 reverse_bits (unsigned char* bits, unsigned long numbits)
 {
     unsigned long numbytes = numbytes_from_numbits (numbits);
-    unsigned char *tmp =
-	(unsigned char*)alloca (numbytes);
+    unsigned char *reverse;
     unsigned long curbit;
     long curbyte = 0;
 	
-    memset (tmp, 0, numbytes);
-
+    reverse = calloc (numbytes, sizeof (unsigned char));
+    if (! reverse) {
+	return huffman_status_memory_failure;
+    }
     for (curbit = 0; curbit < numbits; curbit++) {
 	unsigned int bitpos = curbit % 8;
 
 	if (curbit > 0 && curbit % 8 == 0) {
 	    curbyte++;
 	}
-
-		
-	tmp[curbyte] |= (get_bit (bits, numbits - curbit - 1) << bitpos);
+	reverse[curbyte] |= (get_bit (bits, numbits - curbit - 1) << bitpos);
     }
-
-    memcpy (bits, tmp, numbytes);
+    memcpy (bits, reverse, numbytes);
+    free (reverse);
+    return huffman_status_ok;
 }
 
 /*
  * new_code builds a huffman_code from a leaf in
  * a Huffman tree.
  */
-static huffman_code*
-new_code (const huffman_node* leaf)
+
+/* Build the huffman code by walking up to the root node and then
+  reversing the bits, since the Huffman code is calculated by walking
+  down the tree. */
+
+static huffman_status_t NOIGNORE
+new_code (const huffman_node* leaf, huffman_code ** new_ptr)
 {
-    /* Build the huffman code by walking up to
-     * the root node and then reversing the bits,
-     * since the Huffman code is calculated by
-     * walking down the tree. */
     unsigned long numbits = 0;
-    unsigned char* bits = NULL;
+    unsigned char * bits = NULL;
+    /* Return value. */
     huffman_code *p;
 
     while (leaf && leaf->parent) {
@@ -178,15 +191,15 @@ new_code (const huffman_node* leaf)
 	/* If we need another byte to hold the code,
 	   then allocate it. */
 	if (cur_bit == 0) {
-	    size_t newSize = cur_byte + 1;
+	    size_t new_size = cur_byte + 1;
 	    if (! bits) {
 		MEMPLUS;
 	    }
-	    bits = realloc (bits, newSize);
+	    bits = realloc (bits, new_size);
 	    if (! bits) {
-		return 0;
+		return huffman_status_memory_failure;
 	    }
-	    bits[newSize - 1] = 0; /* Initialize the new byte. */
+	    bits[new_size - 1] = 0; /* Initialize the new byte. */
 	}
 
 	/* If a one must be added then or it in. If a zero
@@ -195,127 +208,151 @@ new_code (const huffman_node* leaf)
 	if (leaf == parent->one) {
 	    bits[cur_byte] |= 1 << cur_bit;
 	}
-
-
 	numbits++;
 	leaf = parent;
     }
 
-    if (bits) {
-	reverse_bits (bits, numbits);
-    }
+    CALL (reverse_bits (bits, numbits));
 
     p = malloc (sizeof (huffman_code));
     if (! p) {
-	return 0;
+	return huffman_status_memory_failure;
     }
     MEMPLUS;
     p->numbits = numbits;
     p->bits = bits;
-    return p;
+    * new_ptr = p;
+    return huffman_status_ok;
 }
 
 /* The maximum number of symbols we allow. */
 
 #define MAX_SYMBOLS 256
-//typedef huffman_node* SymbolFrequencies[MAX_SYMBOLS];
+
+//typedef huffman_node* symbol_frequencies_t[MAX_SYMBOLS];
 //typedef huffman_code* SymbolEncoder[MAX_SYMBOLS];
-typedef huffman_node** SymbolFrequencies;
+//typedef huffman_node** symbol_frequencies_t;
+
+typedef struct symbol_frequencies
+{
+    int n_nodes;
+    huffman_node ** nodes;
+    int allocated;
+}
+symbol_frequencies_t;
+
 typedef huffman_code** SymbolEncoder;
 
 /* Make a new "huffman_node" associated with a symbol. */
 
-static huffman_node*
-new_leaf_node (unsigned char symbol)
+static huffman_status_t
+new_leaf_node (unsigned char symbol, huffman_node ** node_ptr)
 {
     huffman_node *p;
     p = malloc (sizeof (huffman_node));
     if (! p) {
-	return 0;
+	return huffman_status_memory_failure;
     }
     MEMPLUS;
-    p->isLeaf = 1;
+    p->is_leaf = 1;
     p->symbol = symbol;
     p->count = 0;
     p->parent = 0;
-    return p;
+    * node_ptr = p;
+    return huffman_status_ok;
 }
 
 /* Make a new "huffman_node" associated with two child nodes, "zero"
    and "one", without a symbol, but with count "count". */
 
-static huffman_node*
-new_nonleaf_node (unsigned long count, huffman_node *zero, huffman_node *one)
+static huffman_status_t
+new_nonleaf_node (unsigned long count, huffman_node *zero, huffman_node *one,
+		  huffman_node** new_ptr)
 {
     huffman_node *p;
     p = malloc (sizeof (huffman_node));
     if (! p) {
-	return 0;
+	return huffman_status_memory_failure;
     }
     MEMPLUS;
-    p->isLeaf = 0;
+    p->is_leaf = 0;
     p->count = count;
     p->zero = zero;
     p->one = one;
     p->parent = 0;
-	
-    return p;
+    * new_ptr = p;
+    return huffman_status_ok;
 }
 
-static void
+/* Free the memory associated with "subtree". */
+
+static huffman_status_t
 free_huffman_tree (huffman_node *subtree)
 {
     if (subtree == NULL) {
-	return;
+	return huffman_status_ok;
     }
 
-    if (! subtree->isLeaf) {
-	free_huffman_tree (subtree->zero);
-	free_huffman_tree (subtree->one);
+    if (! subtree->is_leaf) {
+	CALL (free_huffman_tree (subtree->zero));
+	subtree->zero = 0;
+	CALL (free_huffman_tree (subtree->one));
+	subtree->one = 0;
     }
 	
     free (subtree);
     MEMFREE;
+    return huffman_status_ok;
 }
 
-static void
+static huffman_status_t
 free_code (huffman_code* p)
 {
     free (p->bits);
     MEMFREE;
     free (p);
     MEMFREE;
+    return huffman_status_ok;
 }
 
-static void
+static huffman_status_t
 free_encoder (SymbolEncoder se)
 {
     unsigned long i;
     for (i = 0; i < MAX_SYMBOLS; i++) {
 	huffman_code *p = se[i];
 	if (p) {
-	    free_code (p);
+	    CALL (free_code (p));
 	}
     }
-
     free (se);
     MEMFREE;
-}
-
-static huffman_status_t NOIGNORE
-init_frequencies (SymbolFrequencies *pSF)
-{
-    *pSF = calloc (MAX_SYMBOLS,  sizeof (huffman_node *));
-    if (! *pSF) {
-	return huffman_status_memory_failure;
-    }
     return huffman_status_ok;
 }
 
 static huffman_status_t NOIGNORE
-free_frequencies (SymbolFrequencies pSF)
+init_frequencies (symbol_frequencies_t ** sf_ptr)
 {
-    free (pSF);
+    symbol_frequencies_t * sf;
+    sf = calloc (1, sizeof (struct symbol_frequencies));
+    if (! sf) {
+	return huffman_status_memory_failure;
+    }
+    // Initially, allocate this many nodes. This can be expanded later.
+    sf->allocated = MAX_SYMBOLS;
+    sf->nodes = calloc (sf->allocated,  sizeof (huffman_node *));
+    if (! sf->nodes) {
+	return huffman_status_memory_failure;
+    }
+    * sf_ptr = sf;
+    return huffman_status_ok;
+}
+
+static huffman_status_t NOIGNORE
+free_frequencies (symbol_frequencies_t * sf)
+{
+    free (sf->nodes);
+    free (sf);
     return huffman_status_ok;
 }
 
@@ -341,8 +378,8 @@ buf_cache;
 static huffman_status_t NOIGNORE
 init_cache (buf_cache* pc,
 	    unsigned int cache_size,
-	    unsigned char **pbufout,
-	    unsigned int *pbufoutlen)
+	    unsigned char ** pbufout,
+	    unsigned int * pbufoutlen)
 {
     if (! pc || ! pbufout || ! pbufoutlen) {
 	return huffman_status_null_pointer;
@@ -363,13 +400,15 @@ init_cache (buf_cache* pc,
     return huffman_status_ok;
 }
 
-static void free_cache (buf_cache* pc)
+static huffman_status_t
+free_cache (buf_cache * pc)
 {
     if (pc->cache) {
 	free (pc->cache);
 	MEMFREE;
 	pc->cache = NULL;
     }
+    return huffman_status_ok;
 }
 
 /* Copy the contents of pc->cache into *pc->bufout, adjusting the
@@ -425,11 +464,7 @@ write_cache (buf_cache * pc,
 
     if (to_write_len > pc->cache_len - pc->cache_cur) {
 	unsigned int newlen;
-	huffman_status_t rc;
-	rc = flush_cache (pc);
-	if (rc != huffman_status_ok) {
-	    return rc;
-	}
+	CALL (flush_cache (pc));
 	newlen = *pc->pbufoutlen + to_write_len;
 	if (! *pc->pbufout) {
 	    MEMPLUS;
@@ -452,25 +487,53 @@ write_cache (buf_cache * pc,
     return huffman_status_ok;
 }
 
-static unsigned int
-get_symbol_frequencies_from_memory (SymbolFrequencies sf,
+static huffman_status_t NOIGNORE
+symbol_frequencies_expand (symbol_frequencies_t * sf)
+{
+    sf->allocated *= 2;
+    sf->nodes = realloc (sf->nodes, sf->allocated);
+    if (! sf->nodes) {
+	return huffman_status_memory_failure;
+    }
+    return huffman_status_ok;
+}
+
+#define UNASSIGNED -1
+
+static huffman_status_t NOIGNORE
+get_symbol_frequencies_from_memory (symbol_frequencies_t * sf,
 				    const unsigned char *bufin,
-				    unsigned int bufinlen)
+				    unsigned int bufinlen,
+				    unsigned int * total_count_ptr)
 {
     unsigned int i;
     unsigned int total_count = 0;
-	
+    int char_to_node[MAX_SYMBOLS];
+    unsigned int n_nodes;
+    for (i = 0; i < MAX_SYMBOLS; i++) {
+	char_to_node[i] = UNASSIGNED;
+    }
+    n_nodes = 0;
     /* Count the frequency of each symbol in the input file. */
     for (i = 0; i < bufinlen; i++) {
 	unsigned char uc = bufin[i];
-	if (! sf[uc]) {
-	    sf[uc] = new_leaf_node (uc);
+	if (char_to_node[uc] == UNASSIGNED) {
+	    CALL (new_leaf_node (uc, & sf->nodes[n_nodes]));
+	    char_to_node[uc] = n_nodes;
+	    n_nodes++;
+	    if (n_nodes > sf->allocated) {
+		CALL (symbol_frequencies_expand (sf));
+	    }
 	}
-	sf[uc]->count++;
+	if (char_to_node[uc] < 0 || char_to_node[uc] >= n_nodes) {
+	    return huffman_status_bounds;
+	}
+	sf->nodes[char_to_node[uc]]->count++;
 	total_count++;
     }
-
-    return total_count;
+    sf->n_nodes = n_nodes;
+    * total_count_ptr = total_count;
+    return huffman_status_ok;
 }
 
 /*
@@ -480,6 +543,7 @@ get_symbol_frequencies_from_memory (SymbolFrequencies sf,
  * the symbol with the lowest frequency is first. Any
  * NULL entries will be sorted to the end of the list.
  */
+
 static int
 SFComp (const void *p1, const void *p2)
 {
@@ -506,12 +570,12 @@ SFComp (const void *p1, const void *p2)
 }
 
 static void UNUSED
-print_freqs (const SymbolFrequencies sf)
+print_freqs (const symbol_frequencies_t * sf)
 {
     size_t i;
-    for (i = 0; i < MAX_SYMBOLS; i++) {
-	if (sf[i]) {
-	    printf ("%d, %ld\n", sf[i]->symbol, sf[i]->count);
+    for (i = 0; i < sf->n_nodes; i++) {
+	if (sf->nodes[i]) {
+	    printf ("%d, %g\n", sf->nodes[i]->symbol, sf->nodes[i]->count);
 	}
 	else {
 	    printf ("NULL\n");
@@ -520,100 +584,90 @@ print_freqs (const SymbolFrequencies sf)
 }
 
 /*
- * build_symbol_encoder builds a SymbolEncoder by walking
- * down to the leaves of the Huffman tree and then,
- * for each leaf, determines its code.
+  "build_symbol_encoder" builds a "SymbolEncoder" by walking down to
+  the leaves of the Huffman tree and then, for each leaf, determining
+  its code.
  */
+
 static huffman_status_t NOIGNORE
 build_symbol_encoder (huffman_node *subtree, SymbolEncoder se)
 {
     if (subtree == NULL) {
 	return huffman_status_null_pointer;
     }
-
-    if (subtree->isLeaf) {
-	se[subtree->symbol] = new_code (subtree);
+    if (subtree->is_leaf) {
+	CALL (new_code (subtree, & se[subtree->symbol]));
     }
     else {
-	huffman_status_t rc;
-	rc = build_symbol_encoder (subtree->zero, se);
-	if (rc != huffman_status_ok) {
-	    return rc;
-	}
-	rc = build_symbol_encoder (subtree->one, se);
-	if (rc != huffman_status_ok) {
-	    return rc;
-	}
+	CALL (build_symbol_encoder (subtree->zero, se));
+	CALL (build_symbol_encoder (subtree->one, se));
     }
     return huffman_status_ok;
 }
 
 /*
- * calculate_huffman_codes turns pSF into an array
+ * calculate_huffman_codes turns sf_ptr into an array
  * with a single entry that is the root of the
  * huffman tree. The return value is a SymbolEncoder,
  * which is an array of huffman codes index by symbol value.
  */
 
-static SymbolEncoder
-calculate_huffman_codes (SymbolFrequencies sf)
+static huffman_status_t
+calculate_huffman_codes (symbol_frequencies_t * sf, SymbolEncoder * se_ptr)
 {
     unsigned int i = 0;
     unsigned int n = 0;
     huffman_node *m1 = NULL, *m2 = NULL;
-    SymbolEncoder pSE;
-    huffman_status_t rc;
+    SymbolEncoder se;
 #if 0
     printf ("BEFORE SORT\n");
     print_freqs (sf);
-#endif
+#endif /* 0 */
 
     /* Sort the symbol frequency array by ascending frequency. */
-    qsort (sf, MAX_SYMBOLS, sizeof (huffman_node *), SFComp);
+    qsort (sf->nodes, sf->n_nodes, sizeof (* sf->nodes), SFComp);
 
 #if 0
     printf ("AFTER SORT\n");
     print_freqs (sf);
-#endif
+#endif /* 0 */
 
     /* Get the number of symbols. */
-    for (n = 0; n < MAX_SYMBOLS && sf[n]; n++) {
+    for (n = 0; n < MAX_SYMBOLS && sf->nodes[n]; n++) {
 	;
     }
 
     /*
-     * Construct a Huffman tree. This code is based
-     * on the algorithm given in Managing Gigabytes
-     * by Ian Witten et al, 2nd edition, page 34.
-     * Note that this implementation uses a simple
-     * count instead of probability.
+      Construct a Huffman tree. This code is based on the algorithm
+      given in Managing Gigabytes by Ian Witten et al, 2nd edition,
+      page 34.  This implementation uses a simple count instead of
+      probability.
      */
     for (i = 0; i < n - 1; i++) {
+	huffman_node * new;
 	/* Set m1 and m2 to the two subsets of least probability. */
-	m1 = sf[0];
-	m2 = sf[1];
+	m1 = sf->nodes[0];
+	m2 = sf->nodes[1];
 
 	/* Replace m1 and m2 with a set {m1, m2} whose probability
 	 * is the sum of that of m1 and m2. */
-	sf[0] = m1->parent = m2->parent =
-	    new_nonleaf_node (m1->count + m2->count, m1, m2);
-	sf[1] = NULL;
+	CALL (new_nonleaf_node (m1->count + m2->count, m1, m2, & new));
+	sf->nodes[0] = m1->parent = m2->parent = new;
+	sf->nodes[1] = NULL;
 		
-	/* Put newSet into the correct count position in pSF. */
-	qsort (sf, n, sizeof (huffman_node *), SFComp);
+	/* Put newSet into the correct count position. */
+	qsort (sf->nodes, n, sizeof (huffman_node *), SFComp);
     }
 
     /* Build the SymbolEncoder array from the tree. */
-    pSE = calloc (MAX_SYMBOLS, sizeof (huffman_code *));
-    if (! pSE) {
+    se = calloc (MAX_SYMBOLS, sizeof (huffman_code *));
+    if (! se) {
 	return 0;
     }
     MEMPLUS;
-    rc = build_symbol_encoder (sf[0], pSE);
-    if (rc != huffman_status_ok) {
-	return 0;
-    }
-    return pSE;
+    CALL (build_symbol_encoder (sf->nodes[0], se));
+    * se_ptr = se;
+    return huffman_status_ok;
 }
 
 /*
@@ -639,8 +693,8 @@ write_code_table_to_memory (buf_cache * pc,
 			    SymbolEncoder se,
 			    uint32_t symbol_count)
 {
-    uint32_t i, count = 0;
-    huffman_status_t rc;
+    uint32_t i;
+    uint32_t count = 0;
 
     /* Determine the number of entries in se. */
     for (i = 0; i < MAX_SYMBOLS; i++) {
@@ -651,16 +705,10 @@ write_code_table_to_memory (buf_cache * pc,
 
     /* Write the number of entries in network byte order. */
     i = htonl (count);
-    rc = write_cache (pc, &i, sizeof (i));
-    if (rc != huffman_status_ok) {
-	return rc;
-    }
+    CALL (write_cache (pc, &i, sizeof (i)));
     /* Write the number of bytes that will be encoded. */
     symbol_count = htonl (symbol_count);
-    rc = write_cache (pc, &symbol_count, sizeof (symbol_count));
-    if (rc != huffman_status_ok) {
-	return rc;
-    }
+    CALL (write_cache (pc, &symbol_count, sizeof (symbol_count)));
 
     /* Write the entries. */
     for (i = 0; i < MAX_SYMBOLS; i++) {
@@ -671,22 +719,13 @@ write_code_table_to_memory (buf_cache * pc,
 	       be stored in an unsigned char. */
 	    unsigned char uc = (unsigned char)i;
 	    /* Write the 1 byte symbol. */
-	    rc = write_cache (pc, &uc, sizeof (uc));
-	    if (rc != huffman_status_ok) {
-		return rc;
-	    }
+	    CALL (write_cache (pc, &uc, sizeof (uc)));
 	    /* Write the 1 byte code bit length. */
 	    uc = (unsigned char)p->numbits;
-	    rc = write_cache (pc, &uc, sizeof (uc));
-	    if (rc != huffman_status_ok) {
-		return rc;
-	    }
+	    CALL (write_cache (pc, &uc, sizeof (uc)));
 	    /* Write the code bytes. */
 	    numbytes = numbytes_from_numbits (p->numbits);
-	    rc = write_cache (pc, p->bits, numbytes);
-	    if (rc != huffman_status_ok) {
-		return rc;
-	    }
+	    CALL (write_cache (pc, p->bits, numbytes));
 	}
     }
     return huffman_status_ok;
@@ -720,33 +759,25 @@ memread (const unsigned char* buf,
 
 /* "pindex" is the offset into bufin. */
 
-static huffman_node*
+static huffman_status_t
 read_code_table_from_memory (const unsigned char* bufin,
 			     unsigned int bufinlen,
 			     unsigned int *pindex,
-			     uint32_t *pDataBytes)
+			     uint32_t *pDataBytes,
+			     huffman_node**root_ptr)
 {
-    huffman_node *root = new_nonleaf_node (0, NULL, NULL);
+    huffman_node *root;
     uint32_t count;
 
-    if (! root) {
-	return NULL;
-    }
+    CALL (new_nonleaf_node (0, NULL, NULL, & root));
 	
     /* Read the number of entries.
        (it is stored in network byte order). */
-    if (memread (bufin, bufinlen, pindex, &count, sizeof (count))) {
-	free_huffman_tree (root);
-	return NULL;
-    }
-
+    CALL (memread (bufin, bufinlen, pindex, &count, sizeof (count)));
     count = ntohl (count);
 
     /* Read the number of data bytes this encoding represents. */
-    if (memread (bufin, bufinlen, pindex, pDataBytes, sizeof (*pDataBytes))) {
-	free_huffman_tree (root);
-	return NULL;
-    }
+    CALL (memread (bufin, bufinlen, pindex, pDataBytes, sizeof (*pDataBytes)));
 
     *pDataBytes = ntohl (*pDataBytes);
 
@@ -759,50 +790,43 @@ read_code_table_from_memory (const unsigned char* bufin,
 	unsigned char *bytes;
 	huffman_node *p = root;
 
-	if (memread (bufin, bufinlen, pindex, &symbol, sizeof (symbol))) {
-	    free_huffman_tree (root);
-	    return NULL;
-	}
-
-	if (memread (bufin, bufinlen, pindex, &numbits, sizeof (numbits))) {
-	    free_huffman_tree (root);
-	    return NULL;
-	}
+	CALL (memread (bufin, bufinlen, pindex, &symbol, sizeof (symbol)));
+	CALL (memread (bufin, bufinlen, pindex, &numbits, sizeof (numbits)));
 		
 	numbytes = (unsigned char) numbytes_from_numbits (numbits);
 	bytes = malloc (numbytes);
 	if (! bytes) {
-	    return NULL;
+	    return huffman_status_memory_failure;
 	}
 	MEMPLUS;
-	if (memread (bufin, bufinlen, pindex, bytes, numbytes)) {
-	    free (bytes);
-	    MEMFREE;
-	    free_huffman_tree (root);
-	    return NULL;
-	}
+	CALL (memread (bufin, bufinlen, pindex, bytes, numbytes));
 
 	/*
-	 * Add the entry to the Huffman tree. The value
-	 * of the current bit is used switch between
-	 * zero and one child nodes in the tree. New nodes
-	 * are added as needed in the tree.
+	 * Add the entry to the Huffman tree. The value of the current
+	 * bit is used to switch between zero and one child nodes in
+	 * the tree. New nodes are added as needed in the tree.
 	 */
 	for (curbit = 0; curbit < numbits; curbit++) {
 	    if (get_bit (bytes, curbit)) {
 		if (p->one == NULL) {
-		    p->one = curbit == (unsigned char)(numbits - 1)
-			? new_leaf_node (symbol)
-			: new_nonleaf_node (0, NULL, NULL);
+		    if (curbit == (unsigned char)(numbits - 1)) {
+			CALL (new_leaf_node (symbol, & p->one));
+		    }
+		    else {
+			CALL (new_nonleaf_node (0, NULL, NULL, & p->one));
+		    }
 		    p->one->parent = p;
 		}
 		p = p->one;
 	    }
 	    else {
 		if (p->zero == NULL) {
-		    p->zero = curbit == (unsigned char)(numbits - 1)
-			? new_leaf_node (symbol)
-			: new_nonleaf_node (0, NULL, NULL);
+		    if (curbit == (unsigned char)(numbits - 1)) {
+			CALL (new_leaf_node (symbol, & p->zero));
+		    }
+		    else {
+			CALL (new_nonleaf_node (0, NULL, NULL, & p->zero));
+		    }
 		    p->zero->parent = p;
 		}
 		p = p->zero;
@@ -813,9 +837,13 @@ read_code_table_from_memory (const unsigned char* bufin,
 	MEMFREE;
     }
 
-    return root;
+    * root_ptr = root;
+    return huffman_status_ok;
 }
 
+/* Encode the memory in "bufin" into "pc" using "se" for the Huffman
+   codes. */
+
 static int
 do_memory_encode (buf_cache *pc,
 		  const unsigned char* bufin,
@@ -873,10 +901,9 @@ huffman_encode_memory (const unsigned char *bufin,
 		       unsigned char **pbufout,
 		       unsigned int *pbufoutlen)
 {
-    SymbolFrequencies sf;
+    symbol_frequencies_t * sf;
     SymbolEncoder se;
     huffman_node *root = NULL;
-    huffman_status_t rc;
     unsigned int symbol_count;
     buf_cache cache;
 
@@ -886,43 +913,30 @@ huffman_encode_memory (const unsigned char *bufin,
 	return huffman_status_null_pointer;
     }
 
-    rc = init_cache (&cache, CACHE_SIZE, pbufout, pbufoutlen);
-    if (rc != huffman_status_ok) {
-	return rc;
-    }
-
+    CALL (init_cache (&cache, CACHE_SIZE, pbufout, pbufoutlen));
     /* Allocate memory for sf and set all its frequencies to 0. */
-    rc = init_frequencies (& sf);
-    if (rc != huffman_status_ok) {
-	return rc;
-    }
-
+    CALL (init_frequencies (& sf));
     /* Get the frequency of each symbol in the input memory. */
-    symbol_count = get_symbol_frequencies_from_memory (sf, bufin, bufinlen);
+    CALL (get_symbol_frequencies_from_memory (sf, bufin, bufinlen,
+					     & symbol_count));
 
     /* Build an optimal table from the symbolCount. */
-    se = calculate_huffman_codes (sf);
-    root = sf[0];
+    CALL (calculate_huffman_codes (sf, & se));
+    root = sf->nodes[0];
 
     /* Scan the memory again and, using the table
        previously built, encode it into the output memory. */
-    rc = write_code_table_to_memory (& cache, se, symbol_count);
-    if (rc == huffman_status_ok) {
-	rc = do_memory_encode (& cache, bufin, bufinlen, se);
-    }
-
+    CALL (write_code_table_to_memory (& cache, se, symbol_count));
+    CALL (do_memory_encode (& cache, bufin, bufinlen, se));
     /* Flush the cache. */
-    rc = flush_cache (& cache);
-    if (rc != huffman_status_ok) {
-	return rc;
-    }
+    CALL (flush_cache (& cache));
     /* Free the Huffman tree. */
-    free_huffman_tree (root);
-    free_encoder (se);
-    free_cache (& cache);
-    rc = free_frequencies (sf);
+    CALL (free_huffman_tree (root));
+    CALL (free_encoder (se));
+    CALL (free_cache (& cache));
+    CALL (free_frequencies (sf));
     MEMCHECK (1);
-    return rc;
+    return huffman_status_ok;
 }
 
 huffman_status_t NOIGNORE
@@ -944,11 +958,8 @@ huffman_decode_memory (const unsigned char *bufin,
     }
 
     /* Read the Huffman code table. */
-    root = read_code_table_from_memory (bufin, bufinlen, &i, &data_count);
-    if (! root) {
-	return huffman_status_null_pointer;
-    }
-
+    CALL (read_code_table_from_memory (bufin, bufinlen, &i, &data_count,
+				       & root));
     buf = malloc (data_count);
     if (! buf) {
 	return huffman_status_memory_failure;
@@ -964,7 +975,7 @@ huffman_decode_memory (const unsigned char *bufin,
 	    p = byte & mask ? p->one : p->zero;
 	    mask <<= 1;
 
-	    if (p->isLeaf) {
+	    if (p->is_leaf) {
 		buf[bufcur] = p->symbol;
 		bufcur++;
 		p = root;
@@ -973,7 +984,7 @@ huffman_decode_memory (const unsigned char *bufin,
 	}
     }
 
-    free_huffman_tree (root);
+    CALL (free_huffman_tree (root));
     *pbufout = buf;
     *pbufoutlen = bufcur;
     MEMCHECK (1);
